(***** This is the model file for the toymodel with one scalar two fermions and a U(1)xSU(3) gauge group *******)

(***************************************************************************************************************)
(******                                        Things to consider:                                        ******)
(******                                                                                                   ******)
(******       1] For every field, every charge has to be specified, even when 0                           ******)
(******       2] For the field strengths and gauge bosons, the following naming scheme has to be applied: ******)
(******          Fieldstrength: name, e.g. "F"                                                            ******)
(******          GaugeBoson: "A" + name, e.g. "AF"                                                        ******)
(******          GaugeBoson: "G" + name, e.g. "AG"                                                        ******)
(******       3] There is also a namingscheme for the indices that has to be upheld:                      ******)
(******          The name of an index belonging to a gauge group has to start with the gauge group name;  ******)
(******          e.g. Groupname = "SU2"  -----> IndexName = "SU2****..."                                  ******)
(******       4] For each pair of Weyl Fermions, one has to specify a Dirac fermion.                      ******)
(******                                                                                                   ******)
(***************************************************************************************************************)

(***** Indices ******)

IndexRange[Index[su3F]] = Unfold[Range[3]];
IndexRange[Index[su3A]] = Unfold[Range[8]];

IndexStyle[ su3F, j];
IndexStyle[ su3A, l];



(***** Parameter list ******)

M$Parameters = {

    gs1 == { 
        ParameterType       -> Internal, 
        TeX                 -> Subscript[g,s]
    },

    gg == { 
        ParameterType       -> Internal, 
        TeX                 -> Subscript[g,g]
    }

}



(***** Gauge group list ******)

M$GaugeGroups = {

    U1Y == {
      Abelian -> True,
      GaugeBoson -> AB,
      Charge -> Y,
      CouplingConstant -> gs1
    },

    su3 == {
        Abelian -> False,
        GaugeBoson -> AG,
        CouplingConstant -> gg,
        StructureConstant -> f,
        Representations -> {{Ta,su3F}},
        FundamentalDimension -> 3
    }
}



(***** Particle classes list ******)

M$ClassesDescription = {

    S[1] == {
        ClassName       -> Phi,
        SelfConjugate   -> True,
        TeXParticleName -> \[Phi],
        QuantumNumbers  -> {Y -> 0}
    },
    
    W[1] == {
        ClassName       -> L1,
        Chirality       -> Left,
        SelfConjugate   -> False,
        TeXParticleName -> Subscript[L,1],
        Indices         -> {Index[su3F]},
        QuantumNumbers  -> {Y -> 1}
    }, 
    
    W[2] == {
        ClassName       -> R1,
        Chirality       -> Right,
        SelfConjugate   -> False,
        TeXParticleName -> Subscript[R,1],
        Indices         -> {Index[su3F]},
        QuantumNumbers  -> {Y -> 1}
    },

    W[3] == {
        ClassName       -> L2,
        Chirality       -> Left,
        SelfConjugate   -> False,
        TeXParticleName -> Subscript[L,2],
        Indices         -> {Index[su3F]},
        QuantumNumbers  -> {Y -> 2}
    },

    W[4] == {
        ClassName       -> R2,
        Chirality       -> Right,
        SelfConjugate   -> False,
        TeXParticleName -> Subscript[R,2],
        Indices         -> {Index[su3F]},
        QuantumNumbers  -> {Y -> 2}
    },

    V[1] == {
        ClassName       -> AB,
        Unphysical      -> True,
        SelfConjugate   -> True,
        TeXParticleName -> AB,
        QuantumNumbers  -> {Y -> 0}
    },

    V[2] == {
        ClassName       -> AG,
        Unphysical      -> True,
        SelfConjugate   -> True,
        TeXParticleName -> AG,
        Indices         -> {Index[su3A]},
        QuantumNumbers  -> {Y -> 0}
    },

    F[1] == {
        ClassName       -> Psi1,
        SelfConjugate   -> False,
        WeylComponents  -> {L1, R1},
	Indices         -> {Index[su3F]},
        QuantumNumbers  -> {Y -> 1}
    },

    F[2] == {
        ClassName       -> Psi2,
        SelfConjugate   -> False,
        WeylComponents  -> {L2, R2}
    }
}

