#This file holds the class definitions for the different classes that are used to read out the
#model file and store them into instances of the container classes


from DEFT import Field, Invariant, Term, Index
import copy as copy
from Utility.readers import Reader
import json


class SigmaPrinter:

    def PrintSigma(self):
        return self.name

    def PrintIndices(self):
        indexstring = ""
        indexstring += self.vectorindex[0] + str(self.vectorindex[1])
        indexstring += ","
        for index in self.indices:
            indexstring += index[0] + str(index[1])
            indexstring += ","
        indexstring = indexstring[:-1]
        

        return indexstring


    def __init__(self,spinindices,vectorindex):
        self.name = "si"
        self.indices = spinindices
        self.vectorindex = vectorindex


class GeneratorPrinter:

    def PrintGenerator(self):
        return self.name

    def AppendIndices(self):
        if self.name.find("[") != -1:
            self.name = self.name[:self.name.find("[")]
        indexstring = ""
        indexstring += self.adjointindex[0] + str(self.adjointindex[1]) + ","
        for index in self.fundamentalindices:
            indexstring += index[0]
            indexstring += str(index[1])
            indexstring += ","
        indexstring = indexstring[:-1]

        self.name += "[" + indexstring + "]"


    def __init__(self,name,fundamentalindices,adjointindex):
        self.name = name
        self.fundamentalindices = fundamentalindices
        self.adjointindex = adjointindex


class SigmaBarPrinter:

    def PrintSigma(self):
        return self.name

    def PrintIndices(self):
        indexstring = ""
        indexstring += self.vectorindex[0] + str(self.vectorindex[1])
        indexstring += ","
        for index in self.indices:
            indexstring += index[0] + str(index[1])
            indexstring += ","
        indexstring = indexstring[:-1]
        

        return indexstring


    def __init__(self,spinindices,vectorindex):
        self.name = "sibar"
        self.indices = spinindices
        self.vectorindex = vectorindex


class InvariantPrinter:

    def PrintInvariant(self):
        return self.name

    def AppendIndices(self):
        if self.name.find("[") != -1:
            self.name = self.name[:self.name.find("[")]
        indexstring = ""
        for index in self.indices:
            indexstring += index[0]
            indexstring += str(index[1])
            indexstring += ","
        indexstring = indexstring[:-1]

        self.name += "[" + indexstring + "]"


    def __init__(self,invariant):
        if invariant.name == 'epsl':
            if 'lor' in invariant.indices[0].group:
                self.name = 'Ueps'
            else:
                self.name = 'Eps'
        elif invariant.name == 'epsu':
            if 'lor' in invariant.indices[0].group:
                self.name = 'Deps'
            else:
                self.name = 'Eps'
        else:
            self.name = invariant.name
        self.indices = []  #these should be triples of the form (indexgroup,label,upper)

        for index in invariant.indices:
            self.indices.append((index.group,index.label,not(index.upper))) #for some reason the upper indices of DEFT invariants have the value "upper = False". Here I choose the convention, that upper indices have "upper = True", dah...

        self.AppendIndices()


class FieldPrinter:

    def PrintField(self):
        fieldstring = ""
        fieldstring += self.name
        return fieldstring

    def AppendIndices(self):
        if self.name.find("[") != -1:
            self.name = self.name[:self.name.find("[")]
        indexstring = ""
        for index in self.indices:
            indexstring += index[0]
            indexstring += str(index[1])
            indexstring += ","
        indexstring = indexstring[:-1]

        self.name += "[" + indexstring + "]"
 

    def __init__(self,inputfield):
        self.field = inputfield._DEFTcopy()
        self.basename = self.field._name
        self.isConjugate = self.field.isConjugate
        

        if self.isConjugate:
            self.name = self.field._name[:-1] + 'bar'
            self.isConjugate = True
        else:
            self.name = self.field._name
            self.isConjugate = False

        self.covariantderivatives = [] #these should be triples of indices, each belonging to one D symbol
        self.indices = []      #these should be triples of the form (indexgroup,label,upper)

        #collect all the indices of field that belong to a covariant derivative
        Dindices = []
        for Dindex in self.field.Dindices:
            Dindices.append(Dindex[0])
            Dindices.append(Dindex[1])

        for index in self.field.indices:
            if index not in Dindices:
                self.indices.append((index.group,index.label,index.upper))

        for cd in self.field.Dindices:
            self.covariantderivatives.append(((cd[0].group,cd[0].label,cd[0].upper),(cd[1].group,cd[1].label,cd[1].upper)))

        for Dindex in Dindices:
            self.field.indices.remove(Dindex)
        self.isDEFTFieldStrength = self.field.isFieldStrength()  # does not return is field strength if there are derivatives attached to it!!!! fixed by above deletion of Dindices from the fields index list
        self.isFeynRulesFieldStrength = False

        #append the indices
        self.AppendIndices()


class FieldStrengthPrinter:

    def generateName(self):
        FSstring = ""
        FSstring += "FS["
        FSstring += self.vectorpotential
        FSstring += ","
        FSstring += self.PrintIndices()
        FSstring += ']'

        return FSstring

    def PrintField(self):
        fieldstring = ""
        fieldstring += self.name


        return fieldstring
        
    def PrintIndices(self):
        indexstring = ""
        for index in self.indices:
            indexstring += index[0]
            indexstring += str(index[1])
            indexstring += ","
        indexstring = indexstring[:-1]

        if self.adjointindex[0] != 'None':
            indexstring += "," + self.adjointindex[0] + str(self.adjointindex[1]) 

        return indexstring


    def __init__(self,field,vecindices,adjointindex):
        if field.isConjugate:
            self.vectorpotential = 'A'+field.basename[:-1]
        else:
            self.vectorpotential = 'A'+field.basename
        self.indices = vecindices
        self.adjointindex = adjointindex
        self.isDEFTFieldStrength = False
        self.isFeynRulesFieldStrength = True

        self.name = self.generateName()

        self.covariantderivatives = field.covariantderivatives


class TermPrinter:

    def TransformFields(self,field):
        newfieldprinter = FieldPrinter(field)
        self.fields.append(newfieldprinter)

    def TransformInvariant(self,invariant):
        newinvariantprinter = InvariantPrinter(invariant)
        self.invariants.append(newinvariantprinter)

    def BuildTermString(self):
        termstring = ""
        for invariant in self.invariants:
            invariantstring = invariant.PrintInvariant()
            termstring += invariantstring
            termstring += ' '
        #now add the fermions with the proper nc environment
        hasfermions = False
        for field in self.fields:
            if not field.isFeynRulesFieldStrength:
                hasfermions = True
        if hasfermions:
            termstring += 'nc['
            for field in self.fields:
                if not field.isFeynRulesFieldStrength:
                    termstring += field.PrintField()
                    termstring += ','
            #close the nc environment (after cutting off the trailing comma)
            termstring = termstring[:-1] + '] '
        #now add the Field Strengths
        for field in self.fields:
            if field.isFeynRulesFieldStrength:
                termstring += field.PrintField()
                termstring += ' '
        for sigma in self.sigmas:
            sigmastring = sigma.PrintSigma()
            termstring += sigmastring
            #indices
            termstring += "["
            indexstring = sigma.PrintIndices()
            termstring += indexstring
            termstring += "] "
        for sigmabar in self.sigmabars:
            sigmastring = sigmabar.PrintSigma()
            termstring += sigmastring
            #indices
            termstring += "["
            indexstring = sigmabar.PrintIndices()
            termstring += indexstring
            termstring += "] "
        for generator in self.generators:
            termstring += generator.PrintGenerator()
            termstring += ' '

        self.termstring = termstring
    
    def PrintTerm(self):
        self.BuildTermString()
        return self.termstring


    def __init__(self,term):
        self.term = term
        self.fields = []
        self.invariants = []
        self.termstring = ""
        self.sigmas = []
        self.sigmabars = []
        self.generators = []

        self.coefficient = ""

        self.overallsign = 1


        for field in self.term._fields:
            self.TransformFields(field)

        for invariant in self.term._invariants:
            self.TransformInvariant(invariant)


class TermsPrinter:


    def ContractDeltas(self,term):
        deltas = [inv for inv in term.invariants if 'del' in inv.name]

        
        for delt in deltas:
            ind1 = delt.indices[0]
            ind2 = delt.indices[1]
            for field in term.fields: #do it for the fields indices
                for i,index in enumerate(field.indices):
                    if ind1[0] == index[0] and ind1[1] == index[1]:
                        field.indices[i] = ind2
                field.AppendIndices()
            for inv in term.invariants: #do it for the epsilons indices
                if 'del' not in inv.name:
                    for j,index in enumerate(inv.indices):
                        if ind1[0] == index[0] and ind1[1] == index[1]:
                            inv.indices[i] = ind2
                inv.AppendIndices()
            
            
            for field in term.fields: #do it for the fields covariant indices
                for i,indices in enumerate(field.covariantderivatives):
                    if ind1[0] == indices[0][0] and ind1[1] == indices[0][1]:
                        field.indices[i][0] = ind2
                    if ind1[0] == indices[1][0] and ind1[1] == indices[1][1]:
                        field.indices[i][1] = ind2
                inv.AppendIndices()

        #remove the deltas from the term
        term.invariants = [inv for inv in term.invariants if 'del' not in inv.name]

    def ContractEpsilons(self,term):
        epsilons = [inv for inv in term.invariants if 'eps' in inv.name and 'lor' in inv.indices[0][0]] #select only the epsilons that have a Lorentz index
        for eps in epsilons:
            #look for fields that the epsilon shares indices with
            fieldpair = []
            fieldpairpositions = []
            for j,f in enumerate(term.fields): 
                for ind in f.indices:
                    if (eps.indices[0][0] == ind[0] and eps.indices[0][1] == ind[1] and not eps.indices[0][2] == ind[2]) or (eps.indices[1][0] == ind[0] and eps.indices[1][1] == ind[1] and not eps.indices[1][2] == ind[2]):
                        fieldpair.append(copy.deepcopy(f))
                        fieldpairpositions.append(j)
                    else:
                        continue
            #look for sigmas (and sigmabars) that the epsilon shares indices with
            sigmapair = []
            sigmapairpositions = []
            for l,si in enumerate(term.sigmas): 
                for ind in si.indices:
                    if (eps.indices[0][0] == ind[0] and eps.indices[0][1] == ind[1] and not eps.indices[0][2] == ind[2]) or (eps.indices[1][0] == ind[0] and eps.indices[1][1] == ind[1] and not eps.indices[1][2] == ind[2]):
                        sigmapair.append(copy.deepcopy(si))
                        sigmapairpositions.append(l)
                    else:
                        continue
            for k,sibar in enumerate(term.sigmabars): 
                for ind in sibar.indices:
                    if (eps.indices[0][0] == ind[0] and eps.indices[0][1] == ind[1] and not eps.indices[0][2] == ind[2]) or (eps.indices[1][0] == ind[0] and eps.indices[1][1] == ind[1] and not eps.indices[1][2] == ind[2]):
                        sigmapair.append(copy.deepcopy(sibar))
                        sigmapairpositions.append(k)
                    else:
                        continue
            #for the case that the epsilon simply contracts with two fields
            if len(fieldpair) == 2: 
                if 'lorR' in eps.indices[0][0]: #in this case we want to lower the index of the first field so as to comply with the convention to contract f_adot f^adot
                    for i,ind in enumerate(fieldpair[0].indices):
                        if eps.indices[0][0] == ind[0] and eps.indices[0][1] == ind[1]: #if the first index is the one being contracted, we have to add a minus sign to the overall factor
                            fieldpair[0].indices[i] = eps.indices[1]
                            term.overallsign = term.overallsign * -1 #add an overall minus sign to the term because the epsilon indices had to be flipped before contraction
                        elif eps.indices[1][0] == ind[0] and eps.indices[1][1] == ind[1]:
                            fieldpair[0].indices[i] = eps.indices[0]
                        else:
                            continue
                elif 'lorL' in eps.indices[0][0]: #in this case we want to lower the index of the second field so as to comply with the convention to contract f^a f_a
                    for i,ind in enumerate(fieldpair[1].indices):
                        if eps.indices[0][0] == ind[0] and eps.indices[0][1] == ind[1]: #if the first index is the one being contracted, we have to add a minus sign to the overall factor
                            fieldpair[1].indices[i] = eps.indices[1]
                            term.overallsign = term.overallsign * -1 #add an overall minus sign to the term because the epsilon indices had to be flipped before contraction
                        elif eps.indices[1][0] == ind[0] and eps.indices[1][1] == ind[1]:
                            fieldpair[1].indices[i] = eps.indices[0]
                        else:
                            continue
                for k,field in enumerate(fieldpair):
                    term.fields[fieldpairpositions[k]] = field
                
                #reappend the new indices to the name of each field

                for field in term.fields:
                    field.AppendIndices()
            #for the case that the epsilon simply connects two si or sibars
            if len(sigmapair) == 2: 
                if 'lorR' in eps.indices[0][0]: #in this case we want to lower the index of the first field so as to comply with the convention to contract f_adot f^adot
                    for i,ind in enumerate(fieldpair[0].indices):
                        if eps.indices[0][0] == ind[0] and eps.indices[0][1] == ind[1]: #if the first index is the one being contracted, we have to add a minus sign to the overall factor
                            fieldpair[0].indices[i] = eps.indices[1]
                            term.overallsign = term.overallsign * -1 #add an overall minus sign to the term because the epsilon indices had to be flipped before contraction
                        elif eps.indices[1][0] == ind[0] and eps.indices[1][1] == ind[1]:
                            fieldpair[0].indices[i] = eps.indices[0]
                        else:
                            continue
                elif 'lorL' in eps.indices[0][0]: #in this case we want to lower the index of the second field so as to comply with the convention to contract f^a f_a
                    for i,ind in enumerate(fieldpair[1].indices):
                        if eps.indices[0][0] == ind[0] and eps.indices[0][1] == ind[1]: #if the first index is the one being contracted, we have to add a minus sign to the overall factor
                            fieldpair[1].indices[i] = eps.indices[1]
                            term.overallsign = term.overallsign * -1 #add an overall minus sign to the term because the epsilon indices had to be flipped before contraction
                        elif eps.indices[1][0] == ind[0] and eps.indices[1][1] == ind[1]:
                            fieldpair[1].indices[i] = eps.indices[0]
                        else:
                            continue
                for k,field in enumerate(fieldpair):
                    term.fields[fieldpairpositions[k]] = field
            
            
            term.invariants.remove(eps)

    def LowerFields(self,term):
        """
        This method serves to lower the spin indices of the fermion fields, as Feynrules is implicitely assuming each
        of the fermion fields' indieces to be a lower one, while DEFT is assuming the opposite. To accomplish that, we simply
        add an appropriate upper epsilon for each field with the right indices
        """

        #first prepare a list with unused index labels
        listofLorLinds = []
        listofLorRinds = []
        for field in term.fields:
            for ind in field.indices:
                if ind[0] == 'lorL':
                    listofLorLinds.append(ind[1])
                elif ind[0] == 'lorR':
                    listofLorRinds.append(ind[1])
        for inv in term.invariants:
            for ind in inv.indices:
                if ind[0] == 'lorL':
                    listofLorLinds.append(ind[1])
                elif ind[0] == 'lorR':
                    listofLorRinds.append(ind[1])
        for si in term.sigmas:
            for ind in si.indices:
                if ind[0] == 'lorL':
                    listofLorLinds.append(ind[1])
                elif ind[0] == 'lorR':
                    listofLorRinds.append(ind[1])
        for sibar in term.sigmabars:
            for ind in sibar.indices:
                if ind[0] == 'lorL':
                    listofLorLinds.append(ind[1])
                elif ind[0] == 'lorR':
                    listofLorRinds.append(ind[1])

        #now create a range of new indexlabels that are free to use

        if len(listofLorLinds) != 0 :
            lorLmax = max(listofLorLinds)
        else:
            lorLmax = 1
        
        if len(listofLorRinds) != 0:
            lorRmax = max(listofLorRinds)
        else:
            lorRmax = 1

        freeLorLindices = list(reversed(range(lorLmax+1,30)))
        freeLorRindices = list(reversed(range(lorRmax+1,30)))

        for f in term.fields:
            if not f.isDEFTFieldStrength:
                for i,ind in enumerate(f.indices):
                    if ind[0] == 'lorL':
                        dummyLorLInd = ('lorL',freeLorLindices.pop(),True)
                        newepsilon = InvariantPrinter(Invariant('epsl','',[Index(ind[0],ind[1],not ind[2]),Index(dummyLorLInd[0],dummyLorLInd[1],not dummyLorLInd[2])]))
                        f.indices[i] = dummyLorLInd
                        f.AppendIndices()
                        term.invariants.append(newepsilon)
                    if ind[0] == 'lorR':
                        dummyLorRInd = ('lorR',freeLorRindices.pop(),True)
                        newepsilon = InvariantPrinter(Invariant('epsl','',[Index(ind[0],ind[1],not ind[2]),Index(dummyLorRInd[0],dummyLorRInd[1],not dummyLorRInd[2])]))
                        f.indices[i] = dummyLorRInd
                        f.AppendIndices()
                        term.invariants.append(newepsilon)

    def TransformDerivatives(self,term):


        listofVecinds = []

        for field in term.fields:
            for ind in field.indices:
                if ind[0] == 'vec':
                    listofVecinds.append(ind[1])
        for si in term.sigmas:
            if si.vectorindex[0] == 'vec':
                listofVecinds.append(si.vectorindex[1])
        for sibar in term.sigmabars:
            if sibar.vectorindex[0] == 'vec':
                listofVecinds.append(sibar.vectorindex[1])

        if len(listofVecinds) != 0:
            vecmax = max(listofVecinds)
        else:
            vecmax = 1

        freeVecIndices = list(reversed(range(vecmax+1,30)))

        usedVecIndices = []
        for f in term.fields:
            for dc in f.covariantderivatives:
                newvectorindexnum = freeVecIndices.pop()
                newvectorindex = ('vec',str(newvectorindexnum))
                newSigma = SigmaPrinter([dc[0],dc[1]],newvectorindex)
                term.sigmas.append(newSigma)
                usedVecIndices.append(newvectorindex)

        #replace the name of the fields that are covariantly differentiated with the correct DC[field,lorindex] expression

        for f in term.fields:
            for dc in f.covariantderivatives:
                DCstring = ""
                index = usedVecIndices.pop()
                DCstring += "DC["+ f.name +"," + index[0] + str(index[1])+ "]"
                f.name = DCstring

    def TransformFieldstrengths(self,term):

        #first list all the indices (or just their number rather) that are already taken in order to build new, unique ones

        listofLorLinds = []
        listofLorRinds = []
        listofVecinds = []
        listofGroupinds = []
        for field in term.fields:
            for ind in field.indices:
                if ind[0] == 'lorL':
                    listofLorLinds.append(ind[1])
                elif ind[0] == 'lorR':
                    listofLorRinds.append(ind[1])
                elif ind[0] == 'vec':
                    listofVecinds.append(ind[1])

        for inv in term.invariants:
            for ind in inv.indices:
                if ind[0] == 'lorL':
                    listofLorLinds.append(ind[1])
                elif ind[0] == 'lorR':
                    listofLorRinds.append(ind[1])
        for si in term.sigmas:
            for ind in si.indices:
                if ind[0] == 'lorL':
                    listofLorLinds.append(ind[1])
                elif ind[0] == 'lorR':
                    listofLorRinds.append(ind[1])
            if si.vectorindex[0] == 'vec':
                listofVecinds.append(si.vectorindex[1])
        for sibar in term.sigmabars:
            for ind in sibar.indices:
                if ind[0] == 'lorL':
                    listofLorLinds.append(ind[1])
                elif ind[0] == 'lorR':
                    listofLorRinds.append(ind[1])
            if sibar.vectorindex[0] == 'vec':
                listofVecinds.append(sibar.vectorindex[1])

        #now create a range of new indexlabels that are free to use

        if len(listofLorLinds) != 0 :
            lorLmax = max(listofLorLinds)
        else:
            lorLmax = 1
        
        if len(listofLorRinds) != 0:
            lorRmax = max(listofLorRinds)
        else:
            lorRmax = 1

        if len(listofVecinds) != 0:
            vecmax = max(listofVecinds)
        else:
            vecmax = 1

        freeLorLindices = list(reversed(range(lorLmax+1,30)))
        freeLorRindices = list(reversed(range(lorRmax+1,30)))
        freeVecindices = list(reversed(range(vecmax+1,30)))

         



        for l,f in enumerate(term.fields):
            if f.isDEFTFieldStrength:
                
                #retrieve the group indices

                groupindices = [ind for ind in f.indices if 'lor' not in ind[0]]

                if groupindices != []:

                    FSgroup = groupindices[0][0]

                    #generate (as above for the other indices) a list of group indices already taken to create new, unique ones

                    listofGroupinds = []
                    for field in term.fields:
                        for ind in field.indices:
                            if ind[0] == FSgroup:
                                listofGroupinds.append(ind[1])

                    for inv in term.invariants:
                        for ind in inv.indices:
                            if ind[0] == FSgroup:
                                listofGroupinds.append(ind[1])

                    for generator in term.generators:
                        for ind in generator.fundamentalindices:
                            if ind[0] == FSgroup:
                                listofGroupinds.append(ind[1])
                        listofGroupinds.append(generator.adjointindex[1])

                    if len(listofGroupinds) != 0 :
                        Groupmax = max(listofGroupinds)
                    else:
                        Groupmax = 1

                    freeGroupindices = list(reversed(range(Groupmax+1,30)))

                    #generate the indices for the generator

                    fundind1 = groupindices[0]
                    fundind2 = groupindices[1]
                    adjointindex = (FSgroup,freeGroupindices.pop(),True)

                    #generate the generator

                    with open("storage/representations.json","r") as read_file:
                        representations = json.load(read_file)
                    
                    generatorname = representations[FSgroup][0][0]
                    
                    newgenerator = GeneratorPrinter(generatorname,[fundind1,fundind2],adjointindex)
                    newgenerator.AppendIndices()

                    term.generators.append(newgenerator)



                else:
                    adjointindex = ('None',0,False)

                if not f.isConjugate:
                    newterm1 = copy.deepcopy(term)
                    newterm2 = copy.deepcopy(term)

                    #build term1
                    Vec1 = ('vec',freeVecindices.pop(),True)
                    Vec2 = ('vec',freeVecindices.pop(),True)

                    dummylorRnum = freeLorRindices.pop()
                    dummylorLnum = freeLorLindices.pop()

                    dummylorRd = ('lorR',dummylorRnum,False)
                    dummylorRu = ('lorR',dummylorRnum,True)
                    dummylorLd = ('lorL',dummylorLnum,False)
                    dummylorLu = ('lorL',dummylorLnum,True)


                    newsigma = SigmaPrinter([f.indices[0],dummylorRd],Vec1)
                    
                    newsigmabar = SigmaBarPrinter([dummylorRu,dummylorLu],Vec2)
                    newepsilon = InvariantPrinter(Invariant('epsu','',[Index(dummylorLd[0],dummylorLd[1],not dummylorLd[2]),Index(f.indices[1][0],f.indices[1][1], not f.indices[1][2])]))

                    newfieldstrength = FieldStrengthPrinter(f,[Vec1,Vec2],adjointindex)

                    #replace old field strength with new one
                    newterm1.fields[l] = newfieldstrength
                    #add all the new invariant symbols
                    newterm1.invariants.append(newepsilon)
                    newterm1.sigmas.append(newsigma)
                    newterm1.sigmabars.append(newsigmabar)
                    #adjust the overall factor
                    newterm1.overallsign = newterm1.overallsign * -0.5

                    #build term2
                    Vec1 = ('vec',freeVecindices.pop(),True)
                    Vec2 = ('vec',freeVecindices.pop(),True)

                    dummylorRnum = freeLorRindices.pop()
                    dummylorLnum = freeLorLindices.pop()

                    dummylorRd = ('lorR',dummylorRnum,False)
                    dummylorRu = ('lorR',dummylorRnum,True)
                    dummylorLd = ('lorL',dummylorLnum,False)
                    dummylorLu = ('lorL',dummylorLnum,True)


                    newsigma = SigmaPrinter([f.indices[0],dummylorRd],Vec2)
                    newsigmabar = SigmaBarPrinter([dummylorRu,dummylorLu],Vec1)
                    newepsilon = InvariantPrinter(Invariant('epsu','',[Index(dummylorLd[0],dummylorLd[1],not dummylorLd[2]),Index(f.indices[1][0],f.indices[1][1], not f.indices[1][2])]))

                    newfieldstrength = FieldStrengthPrinter(f,[Vec1,Vec2],adjointindex)

                    #replace old field strength with new one
                    newterm2.fields[l] = newfieldstrength
                    #add all the new invariant symbols
                    newterm2.invariants.append(newepsilon)
                    newterm2.sigmas.append(newsigma)
                    newterm2.sigmabars.append(newsigmabar)
                    #adjust the overall factor
                    newterm2.overallsign = newterm2.overallsign * 0.5
                elif f.isConjugate:
                    newterm1 = copy.deepcopy(term)
                    newterm2 = copy.deepcopy(term)

                    #build term1
                    Vec1 = ('vec',freeVecindices.pop(),True)
                    Vec2 = ('vec',freeVecindices.pop(),True)

                    dummylorRnum = freeLorRindices.pop()
                    dummylorLnum = freeLorLindices.pop()

                    dummylorRd = ('lorR',dummylorRnum,False)
                    dummylorRu = ('lorR',dummylorRnum,True)
                    dummylorLd = ('lorL',dummylorLnum,False)
                    dummylorLu = ('lorL',dummylorLnum,True)


                    newsigma = SigmaPrinter([dummylorLd,f.indices[1]],Vec2)
                    newsigmabar = SigmaBarPrinter([dummylorRu,dummylorLu],Vec1)
                    newepsilon = InvariantPrinter(Invariant('epsu','',[Index(f.indices[0][0],f.indices[0][1],not f.indices[0][2]),Index(dummylorRd[0],dummylorRd[1],not dummylorRd[2])]))

                    newfieldstrength = FieldStrengthPrinter(f,[Vec1,Vec2],adjointindex)

                    #replace old field strength with new one
                    newterm1.fields[l] = newfieldstrength
                    #add all the new invariant symbols
                    newterm1.invariants.append(newepsilon)
                    newterm1.sigmas.append(newsigma)
                    newterm1.sigmabars.append(newsigmabar)
                    #adjust the overall factor
                    newterm1.overallsign = newterm1.overallsign * -0.5

                    #build term2
                    Vec1 = ('vec',freeVecindices.pop(),True)
                    Vec2 = ('vec',freeVecindices.pop(),True)

                    dummylorRnum = freeLorRindices.pop()
                    dummylorLnum = freeLorLindices.pop()

                    dummylorRd = ('lorR',dummylorRnum,False)
                    dummylorRu = ('lorR',dummylorRnum,True)
                    dummylorLd = ('lorL',dummylorLnum,False)
                    dummylorLu = ('lorL',dummylorLnum,True)


                    newsigma = SigmaPrinter([dummylorLd,f.indices[1]],Vec1)
                    newsigmabar = SigmaBarPrinter([dummylorRu,dummylorLu],Vec2)
                    newepsilon = InvariantPrinter(Invariant('epsu','',[Index(f.indices[0][0],f.indices[0][1],not f.indices[0][2]),Index(dummylorRd[0],dummylorRd[1],not dummylorRd[2])]))

                    newfieldstrength = FieldStrengthPrinter(f,[Vec1,Vec2],adjointindex)

                    #replace old field strength with new one
                    newterm2.fields[l] = newfieldstrength
                    #add all the new invariant symbols
                    newterm2.invariants.append(newepsilon)
                    newterm2.sigmas.append(newsigma)
                    newterm2.sigmabars.append(newsigmabar)
                    #adjust the overall factor
                    newterm2.overallsign = newterm2.overallsign * 0.5
                break
        
        return newterm1,newterm2

    def PrintTerms(self):

        termsstring = ""
        for term in self.terms:
            termsstring += str(term.overallsign) + " * " + term.PrintTerm()
            termsstring += "+ "
        
        termsstring = termsstring[:-2] #remove the trailing + symbol
        
        return termsstring

    #init statement

    def __init__(self,seed):
        self.seed = seed
        self.terms = []

        self.terms.append(TermPrinter(self.seed))

        for term in self.terms:
            self.ContractDeltas(term)
        done = False
        while not done:
            for term in self.terms:
                if sum(1 for f in term.fields if f.isDEFTFieldStrength ) != 0:
                    newterm1,newterm2 = self.TransformFieldstrengths(term)
                    self.terms.append(newterm1)
                    self.terms.append(newterm2)
                    self.terms.remove(term)
            done = True
            for term in self.terms: #check if there are fieldstrengths left
                if sum(1 for f in term.fields if f.isDEFTFieldStrength ) != 0:
                    done = False
                    break
                else:
                    continue
        for term in self.terms:
            self.LowerFields(term)
        for term in self.terms:
            self.TransformDerivatives(term)


def UpdateModelFile(Lagrangian,LagrangianCoeffs,dimNCoeffs,modelfilepath):
        

        #open the model file and write

        newReader = Reader()

        newReader.ReadModelFilewSpaces(modelfilepath)

        oldparamdefstring = newReader.ExtractDefinitions('M$Parameters')

        newparamdefstring = oldparamdefstring[:-1] + ",\n\n"
        for coeff in LagrangianCoeffs:
            if coeff != '0':
                newparamdefstring += "    "
                newparamdefstring += coeff
                newparamdefstring += " == {ParameterType        -> External \n    }, \n\n"
        for coeff in dimNCoeffs:
            if coeff != '0':
                newparamdefstring += "    "
                newparamdefstring += coeff
                newparamdefstring += " == {ParameterType        -> External \n    }, \n\n"
        newparamdefstring = 'M$Parameters = {' + newparamdefstring[:-4] + "\n\n"

        newReader.ReplaceDefinitions('M$Parameters',newparamdefstring)

        #append the Lagrangian

        newReader.ModelFileText = newReader.ModelFileText + "\n\n" + Lagrangian

        newReader.WriteModelFile(modelfilepath)


