import sys
sys.path.insert(0,'..')

import os


def convertfrac(expression):
    if "/" in expression:
        slashposition = expression.index("/")
        numerator = expression[:slashposition]
        denominator = expression[(slashposition+1):]
    else:
        numerator = expression
        denominator = "1"
    
    return "frac(" + numerator +","+denominator+")"


class FieldClassDeclaration:

    def __init__(self,classname,dimension,nonabel,uones,name,basename,texname,isSelfConjugate,isU1FieldStrength= False,vectorpotential = None):
        self.classname = classname
        self.dimension= convertfrac(dimension)
        self.NADict = nonabel
        self.U1Dict = uones
        self._name = name
        self._basename = basename
        self.texname = texname
        self.SelfConjugate= isSelfConjugate
        self.U1FieldStrength = isU1FieldStrength
        self.vectorPotential = vectorpotential

        

    def writeDeclarationToFile(self,filepath):
        """
        Description: This method is used to write the finished field class declaration to the appropriate file
        Input: filepath of the file to which the field declaration needs to be written.
        """

        #prepare the text for the field file
        headerline = "class " + self.classname + "(Field):"
        basenameline = "_basename=\'" + self._basename + "\'"
        defline = "def __init__(self,term=None):"

        declarationline = "Field.__init__(self," + self.dimension + ",{"
        for NAgroup in self.NADict.keys():
            declarationline += "\'" + NAgroup + "\':"
            declarationline += "\'" + self.NADict[NAgroup] + "\',"
        declarationline = declarationline[:-1] #get rid of the final comma
        declarationline += "},{"

        for U1group in self.U1Dict.keys():
            declarationline += "\'" + U1group + "\':"
            declarationline +=  self.U1Dict[U1group]+ ","
        declarationline = declarationline[:-1] #get rid of the final comma
        declarationline += "},\'"
        
        declarationline +=  self._name + "\',r\'" + self.texname + "\',term)"


        selfconjugateline = "self.isSelfConjugate=True"

        if self.U1FieldStrength:
            funcdiffdictline = "self.funcDiffDict[\'" + self.vectorPotential + "\']=\'self._FDFS(field)\'"
        



        filetext = headerline + "\n"  + "  " + basenameline + "\n"  + "  "
        
        if self.vectorPotential is not None:
            vectorpotentialline = "_vectorPotential=" + self.vectorPotential
            filetext += vectorpotentialline + "\n"  + "  "
        
        filetext += defline + "\n"  + "    " + declarationline + "\n"

        if self.SelfConjugate and self.vectorPotential is None:
            filetext += "    " + selfconjugateline + "\n"
        if self.U1FieldStrength:
            filetext += "    " + funcdiffdictline + "\n"

        Fieldfile =open(filepath,'a')
        Fieldfile.write(filetext+'\n')
        Fieldfile.close()

