#import necessary modules
import sympy
import random
import itertools
import json
import easygui as egui

#####################################################################################################################
#####################################################################################################################
#                                               TOGGLE MENU                                                         #
#####################################################################################################################
CleanUP = 1                  #--------> should be turned on to clean the storage files at the the beginning         #
#                                       (can be turned off for diagnostic reasons)                                  #
#                                                                                                                   #
GenerateLagrangianTerms = 1  #--------> generate the dimension 4 terms                                              #
#                                       (needs to be turned on when wanting to generate the basis)                  #
#                                                                                                                   #
SpecifyCoefficients = 1      #--------> ask the user to specifiy the coefficients of the Lagrangian terms.          #
#                                       Attention: !!Not specifying them can lead to a wrong dim N basis!!          #                                     
#                                                                                                                   #
GeneratedimNBasis = 1        #--------> generates all the possible terms of dimension d = n                         #                                     
#                                                                                                                   #
#                                                                                                                   #
DoCustomReport = 1           #--------> generates a different report that can be costumized right in this file      #                                    
#                                                                                                                   #
#                                                                                                                   #
generateFeynRulesInput = 1   #--------> Update the Modelfile with the parameterslist and the Lagrangian             #                                        
#                                                                                                                   #
#                                                                                                                   #
#########################   Desired Energy Dimension d = n of the Operator Basis ####################################
#                                                         |                                                         #                          
n = 6  #<-------------------------------------------------|                                                         #           
#                                                                                                                   #                              
#####################################################################################################################





########################################### MODEL FILE NAME AND PATH ################################################

modelpath = "Feynrules/Models/"

filename = "ScalarToymodel_withNA.fr"

modelfile = modelpath + filename

#####################################################################################################################





#################################################### CLEAN UP #######################################################
# Reset the json files as well as the fields file for a clean next run

if CleanUP:

    U1GaugeGroupsCouplingCLEARED = {}
    NAGaugeGroupsCouplingCLEARED = {}
    gaugeGroupsFieldsCLEARED = {}
    groupdefsCLEARED = {"lorL": {"1": (0, 0, []), "2l": (0, 1, []), "3": (0, 2, [("Symmetric", [0, 1])]), "2": (1, 0, [])}, "lorR": {"1": (0, 0, []), "2l": (0, 1, []), "3": (0, 2, [("Symmetric", [0, 1])]), "2": (1, 0, [])}}
    groupsdimfundCLEARED = {"lorL":2,"lorR":2}
    TeXindicesCLEARED = {'lorL':[r'\alpha',r'\beta',r'\gamma',r'\delta',r'\mu',r'\nu',r'\rho',r'\sigma'],'lorR':[r'\dot\alpha',r'\dot\beta',r'\dot\gamma',r'\dot\delta',r'\dot\mu',r'\dot\nu',r'\dot\rho',r'\dot\sigma']}
    representationsCLEARED = {}

    with open("storage/groupcouplings.json", "w") as write_file:
        json.dump([U1GaugeGroupsCouplingCLEARED,NAGaugeGroupsCouplingCLEARED],write_file)
    with open("storage/gaugegroupfields.json", "w") as write_file:
        json.dump(gaugeGroupsFieldsCLEARED,write_file)
    with open("storage/groupdefs.json", "w") as write_file:
        json.dump(groupdefsCLEARED,write_file)
    with open("DEFT/fields.py", "w") as write_file:
        write_file.write("")
    with open("storage/groupsdimfund.json", "w") as write_file:
        json.dump(groupsdimfundCLEARED,write_file)
    with open("storage/TeXindices.json", "w") as write_file:
        json.dump(TeXindicesCLEARED,write_file)
    with open("storage/representations.json", "w") as write_file:
        json.dump(representationsCLEARED,write_file)
    
#####################################################################################################################





#####################################################################################################################

#########  ########  ########  #      #  ########  ########
#          #      #  #      #  #      #  #      #  #
#     ###  ########  #      #  #      #  ########  ########
#       #  #    #    #      #  #      #  #                #
#########  #      #  ########  ########  #         ########

#####################################################################################################################

#import the Reader and Container classes as well as the class that handles the creation of the DEFT fields file
from Utility.readers import Reader, FieldsReader, GaugeGroupsReader, IndicesReader
from Utility.containers import Container, FieldsContainer, GaugeGroupsContainer
from Utility.utility import FieldClassDeclaration




#Load the model file and extract the gauge group information from the model file
newgroupsreader = GaugeGroupsReader()
newgroupsreader.ReadModelFilewoSpaces(modelfile) #read the model file
fullgroupsdefinition = newgroupsreader.ExtractDefinitions("M$GaugeGroups") #separate the group definitions
groupdefinitiondicts = newgroupsreader.ExtractGroupDefinitions(fullgroupsdefinition) # store the group definitions


#Read and extract the index information from the model file
newindicesreader = IndicesReader()
newindicesreader.ReadModelFilewoSpaces(modelfile) #read the model file
indicesdefinitiondicts = newindicesreader.ExtractIndicesDefinitions() # store the indices definitions
 

#Transform the dictionary information and write the gauge information to the json files that subsequently 
#are read by the DEFT routine
groupcontainers = []
representationnames = {} #dictionary relating indexnames to the repnames; {'indexname':('groupname','repname'),...}
NAgaugebosons = {} #make a list with nonabel gauge bosons and their group to know which representations need modified
i = 1

for defdict in groupdefinitiondicts:
    #only select the indices that are relevant for this group
    relevantindicesdefinitiondicts = {}
    groupname = defdict['GroupName']
    for key in indicesdefinitiondicts.keys():
      if groupname in key:
        relevantindicesdefinitiondicts.update({key:indicesdefinitiondicts[key]})
    nc = GaugeGroupsContainer(defdict,relevantindicesdefinitiondicts)
    #initiate a container class instance (transformation of dictionariy information is done right at the initialization)
    representationnames.update(nc.repnames) 
    NAgaugebosons.update({nc.gaugeboson:(nc.groupname,nc.isAbelian)}) 
    groupcontainers.append(nc)
    i += 1


#load the four json files that store the group couplings, the gauge group/gauge field allocation and the groupdefs
with open("storage/groupcouplings.json", "r") as read_file:
    U1GaugeGroupsCoupling,NAGaugeGroupsCoupling= json.load(read_file)
with open("storage/gaugegroupfields.json", "r") as read_file:
    gaugeGroupsFields= json.load(read_file)
with open("storage/groupdefs.json", "r") as read_file:
    groupdefs= json.load(read_file)
with open("storage/groupsdimfund.json", "r") as read_file:
    groupsdimfund= json.load(read_file)
with open("storage/TeXindices.json", "r") as read_file:
    TeXindices= json.load(read_file)
with open("storage/representations.json","r") as read_file:
    representations = json.load(read_file)


#update the dictionaries
for nc in groupcontainers:
    U1GaugeGroupsCoupling.update(nc.U1GaugeGroupCoupling)
    NAGaugeGroupsCoupling.update(nc.NAGaugeGroupCoupling)
    gaugeGroupsFields.update(nc.GaugeGroupField)
    groupdefs.update(nc.groupdef)
    groupsdimfund.update(nc.groupsdimfund)
    TeXindices.update(nc.TeXindices)
    representations.update(nc.representations)

#store the four updated json fields again
with open("storage/groupcouplings.json", "w") as write_file:
    json.dump([U1GaugeGroupsCoupling,NAGaugeGroupsCoupling],write_file)
with open("storage/gaugegroupfields.json", "w") as write_file:
    json.dump(gaugeGroupsFields,write_file)
with open("storage/groupdefs.json", "w") as write_file:
    json.dump(groupdefs,write_file)
with open("storage/groupsdimfund.json", "w") as write_file:
    json.dump(groupsdimfund,write_file)
with open("storage/TeXindices.json", "w") as write_file:
    json.dump(TeXindices,write_file)
with open("storage/representations.json", "w") as write_file:
    json.dump(representations,write_file)

#####################################################################################################################





#####################################################################################################################

########  #   #######  #         #####   #######
#         #   #        #         #    #  #
#####     #   ######   #         #    #  #######
#         #   #        #         #    #        #
#         #   #######  ########  #####   #######

#####################################################################################################################



#Load the model file and extract the field definitions
newfieldsreader = FieldsReader()
newfieldsreader.ReadModelFilewoSpaces(modelfile) #read model file
fullfieldsdefinition = newfieldsreader.ExtractDefinitions("M$ClassesDescription") #separate the field definitions
fielddefinitiondicts = newfieldsreader.ExtractFieldDefinitions(fullfieldsdefinition) # list of dictionaries
#exlcude the dummy fields used by Feynrules for the WeylToDirac conversion
fielddefinitiondicts = [fdict for fdict in fielddefinitiondicts if 'Dummy' not in fdict['ClassName']]


#Transform the information in the field definition dictionaries using the Container Class and build the DEFT fields file

#prepare the DEFT fields file
filetext = "from DEFT import Field,frac,Unitary"
Fieldfile =open("./DEFT/fields.py",'a')
Fieldfile.write(filetext+'\n\n\n')
Fieldfile.close()


fieldcontainers = []
for defdict in fielddefinitiondicts:
    #exclude the fermion declarations
    if defdict['FieldType'] != 'F':
      nc = FieldsContainer(defdict) #initiate a container class instance
      fieldcontainers.append(nc) #store the container instance for later use


#add the representation information to the fields
for group in groupcontainers:
    if not group.isAbelian:
        for field in fieldcontainers:
            field.BuildNonAbels(group.groupname,representationnames[group.groupname])


#swap the representations for the gauge bosons to the correct one and generate the field strength information
FSdicts = []
for field in fieldcontainers:
  if field.name in NAgaugebosons:
    if not NAgaugebosons[field.name][1]:
      field.nonabel[NAgaugebosons[field.name][0]] = 'FS'
    FSdict = field.BuildFieldStrength(NAgaugebosons[field.name][0],NAgaugebosons[field.name][1])
    FSdicts.append(FSdict)



#build the fields file
for nc in fieldcontainers:
    newfieldclassdeclaration = FieldClassDeclaration(nc.classname,nc.dimension,nc.nonabel,nc.uones,nc.name,nc.basename,nc.texname,nc.isSelfConjugate) #initiate the fields file building class instance
    newfieldclassdeclaration.writeDeclarationToFile("./DEFT/fields.py") #write the fields file to the given path

#add the field strengths
for FS in FSdicts:
    newfieldclassdeclaration = FieldClassDeclaration(FS['cn'],FS['dim'],FS['nonabel'],FS['uones'],FS['n'],FS['bn'],FS['tn'],FS['sc'],FS['UoneFS'],FS['vp']) #initiate the fields file building class instance
    newfieldclassdeclaration.writeDeclarationToFile("./DEFT/fields.py") #write the fields file to the given path

#####################################################################################################################





#####################################################################################################################

#####   ######  #######  #########
#    #  #       #            #
#    #  ####    #####        #
#    #  #       #            #
#####   ######  #            #

#####################################################################################################################


#import DEFT modules
import DEFT.fields
import DEFT.operators as op
import DEFT
from DEFT import frac

#build the fieldslist
DEFTFields = []
for nc in fieldcontainers:
    if nc.name[0] == 'A':
      DEFTFields.append(eval("DEFT.fields." + nc.name[1:]))
    else: #this line makes sure that the the field strength is considered instead of the vector potential
        DEFTFields.append(eval("DEFT.fields." + nc.name))


#generate Lagrangian Terms
if GenerateLagrangianTerms:
  dimFourTerms = op.generateAllTerms(DEFTFields,op.dimLEQ(4))
  dimFourTerms = [t for t in dimFourTerms if t.dimension==frac(4,1)]
  LagrangianTerms = dimFourTerms
else:
  LagrangianTerms = []


#Choose the desired Lagrangian terms and coefficients
if SpecifyCoefficients:

  ####### print the dimension 4 Lagrangian and the corresponding shortstrings for identification
    latexstring1 = r'''Lagrangian \\'''
    for i,term in enumerate(LagrangianTerms):
      latexstring1 += "%d) " %(i+1)
      latexstring1 += term.texRep()
      latexstring1 += r'''$\quad \rightarrow \quad$ '''
      latexstring1 += term.shortstr()
      latexstring1 += r'''\\'''
      latexstring1 += '\n'
    op.makePDFDoc(latexstring1,"TermsAndShortstrs")

    termSubs = []
    subs = []
    msg = "Choose the Lagrangian Coefficients."
    title = "Choose Lagrangian"
    fieldNames = [term.shortstr() for term in LagrangianTerms]
    fieldValues = []  # we start with blanks for the values

    #ask in batches of 10 for better visibility
    j = 0
    while True:
      if len(fieldNames) < 10:
        if len(fieldNames) == 0: break
        lastbatchoffieldValues = egui.multenterbox(msg,title, fieldNames)
        if lastbatchoffieldValues == None:
          fieldValues = []
          break
        fieldValues = fieldValues + lastbatchoffieldValues
        break
      if j  == 10:
        batchoffieldValues = egui.multenterbox(msg,title, fieldNames[:j])
        if batchoffieldValues == None:
          fieldValues = []
          break
        fieldNames = fieldNames[j:]
        fieldValues = fieldValues + batchoffieldValues
        j = 0
      j += 1

    #only prepare the substitution coefficients if fieldsvalues list is not emptly

    if fieldValues != []:
      LagrangianCoeffs = op.makeCoeffs(LagrangianTerms)
      for i,term in enumerate(LagrangianTerms):
        if fieldValues[i] != "":
          termSubs.append((term,sympy.sympify(fieldValues[i])))
          subs.append((sympy.sympify(LagrangianCoeffs[i]),sympy.sympify(fieldValues[i])))

else:
    termSubs = []
    subs = []
    fieldValues = []



#generate dimension d = n terms
if GeneratedimNBasis:

  #check first if generateLagrangianTerm was activeted
  if not GenerateLagrangianTerms:
    raise Exception("You cannot generate a correct basis without having generated the d = 4 Lagrangian!")

  dimNTerms=op.generateAllTerms(DEFTFields,op.dimLEQ(n))
  dimNTerms=[t for t in dimNTerms if t.dimension==frac(n,1)]

  #generate relations
  EOMCoeffSubs=[]
  if n>4:  
    LagrangianTermsSimple=op.removeDupsFierzSym(LagrangianTerms)
    LagrangianCoeffsSimple=op.makeCoeffs(LagrangianTermsSimple)
    fdFields=op.makeFDFields(DEFTFields)
    print("***********make FD fields done*************")
    eomrels=op.makeFuncDiffandEOMRelations(fdFields,LagrangianTermsSimple,LagrangianCoeffsSimple,dimNTerms)
    for t,w in zip(LagrangianTermsSimple,LagrangianCoeffsSimple):
      matcht,sub=next(((t2,w2) for t2,w2 in termSubs if t==t2),(None,None))
      if matcht is not None: EOMCoeffSubs.append( (sympy.sympify(w),sub*matcht.compareSign(t)) )
  else: eomrels=[]
  print("***********EOM Relations done*************")
  frs=op.makeFierzRelations(dimNTerms)
  print("***********Fierz Relations done*************")
  sds=op.makeSwitchDerivativeRelations(dimNTerms)
  print("***********switchderivatives Relations done*************")
  ibps=op.makeIBPRelations(dimNTerms)
  print("***********IBP Relations done*************")
  relations=frs+ibps+sds+eomrels


  # Generate extra substitutions to speed up the Row reduction process if so desired
  allCoeffs=set(itertools.chain.from_iterable(itertools.chain.from_iterable(w.free_symbols for w in r.weights) for r in relations))
  subbedCoeffs=[s[0] for s in subs]
  remainingCoeffs=list(allCoeffs-set(subbedCoeffs))
  listOfInts=[23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97, 101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 151, 157, 163,167,173,179,181,191,193,197,199,211,223,227,229,233,239,241,251,257,263,269,271,277,281,283,293,307,311,313,317,331,337,347,349,353,359,367,373,379,383,389,397,401,409,419,421,431,433,439,443,449,457,461,463,467,479,487,491,499,503,509,521,523,541,547,557,563,569,571,577,587,593,599,601,607,613,617,619,631,641,643,647,653,659,661,673,677,683,691,701,709,719,727,733,739,743,751,757,761,769,773,787,797,809,811,821,823,827,829,839,853,857,859,863,877,881,883,887,907,911,919,929,937,941,947,953,967,971,977,983,991,997,1009,1013,1019,1021,1031,1033,1039,1049,1051,1061,1063,1069,1087,1091,1093,1097,1103,1109,1117,1123,1129,1151,1153,1163,1171,1181,1187,1193,1201,1213,1217,1223,1229,1231,1237,1249,1259,1277,1279,1283,1289,1291,1297,1301,1303,1307,1319,1321,1327,1361,1367,1373,1381,1399,1409,1423,1427,1429,1433,1439,1447,1451,1453,1459,1471,1481,1483,1487,1489,1493,1499,1511,1523,1531,1543,1549,1553,1559,1567,1571,1579,1583,1597,1601,1607,1609,1613,1619,1621,1627,1637,1657,1663,1667,1669,1693,1697,1699,1709,1721,1723,1733,1741,1747,1753,1759,1777,1783,1787,1789,1801,1811,1823,1831,1847,1861,1867,1871,1873,1877,1879,1889,1901,1907,1913,1931,1933,1949,1951,1973,1979,1987,1993,1997,1999,2003,2011,2017,2027,2029,2039,2053,2063,2069,2081,2083,2087,2089]
  if len(remainingCoeffs)>len(listOfInts): raise Exception('I don\'t know enough primes!')
  random.shuffle(listOfInts)
  extraSubs=[(co,sympy.Integer(pr)) for co,pr in zip(remainingCoeffs,listOfInts[:len(remainingCoeffs)])]


  # Perform substitutions
  for r in relations:
    for i,w in enumerate(r.weights): r.weights[i]=sympy.simplify(w).subs(EOMCoeffSubs+extraSubs)
  print("***********SUBS done*************")

  # Generate Basis List
  dimNTermss,rrefrls = op.betterTermEliminator(dimNTerms,relations)


  #do analysis
  basisterms = op.analyse(dimNTermss,rrefrls,"Analysis")

else:
  basisterms = []

from Utility.printers import TermsPrinter, UpdateModelFile


#####################################################################################################################

######  #    #  #######  ######## #    # #######
#    #  #    #     #     #      # #    #    #
#    #  #    #     #     ######## #    #    #
#    #  #    #     #     #        #    #    #
######  ######     #     #        ######    #

#####################################################################################################################

#first prepare the coefficients and make sure the declaration from the manual specification
#of the Lagrangian Coeffs get considered

printableLagrangianCoeffs = op.makeCoeffs(LagrangianTerms)
if SpecifyCoefficients:
  for j, value in enumerate(fieldValues):
    if value != "":
      printableLagrangianCoeffs[j] = value
printabledimNCoeffs = op.makeCoeffs(basisterms)


#generate a custom report listing all the terms and the coefficients that have been generated (or choosen) for them
if DoCustomReport:

  ####### print the dimension 4 Lagrangian and the corresponding coefficients
  latexstring = r'''Lagrangian \\'''
  for i,term in enumerate(LagrangianTerms):
    newTermsPrinter = TermsPrinter(term)
    latexstring += "%d) " %(i+1)
    latexstring += term.texRep()
    latexstring += r'''$\quad \rightarrow \quad$ '''
    latexstring += printableLagrangianCoeffs[i]
    latexstring += r'''\\'''
    latexstring += '\n'

  ####### print the dim = n basisterms generated by DEFT
  latexstring += r'''\newpage Terms \\'''
  for i,term in enumerate(basisterms):
    newTermsPrinter = TermsPrinter(term)
    latexstring += "%d) " %(i+1)
    latexstring += term.texRep()
    latexstring += r'''$\quad \rightarrow \quad$ '''
    latexstring += printabledimNCoeffs[i]
    latexstring += r'''\\'''
    latexstring += '\n'
  op.makePDFDoc(latexstring,"TermsAndCoefficients")

#generate the Output for Feynrules
if generateFeynRulesInput:

  #generate the Lagrangian
  Lagrangian = "L = "

  for i,term in enumerate(LagrangianTerms):
    if printableLagrangianCoeffs[i] != '0':
      newTermsPrinter = TermsPrinter(term)
      Lagrangian += " "
      Lagrangian += "ToGrassmannBasis["
      Lagrangian += printableLagrangianCoeffs[i] + " * ("
      Lagrangian += newTermsPrinter.PrintTerms()
      Lagrangian += ")] + "

  Lagrangian = Lagrangian[:-2]  

  LagrangiandimN = ""

  for i,term in enumerate(basisterms):
      newTermsPrinter = TermsPrinter(term)
      LagrangiandimN += " "
      LagrangiandimN += "ToGrassmannBasis["
      LagrangiandimN += printabledimNCoeffs[i] + " * ("
      LagrangiandimN += newTermsPrinter.PrintTerms()
      LagrangiandimN += ")] + "
  
  LagrangiandimN = LagrangiandimN[:-2]

  if LagrangiandimN != "":
    plus = "+"
  else:
    plus = ""

  LagrangianTot = Lagrangian + plus + LagrangiandimN

  #write to modelfile
  UpdateModelFile(LagrangianTot,printableLagrangianCoeffs,printabledimNCoeffs,modelfile)


  









